///////////////////////////////////////// --o-
//
// p5udpinput.js
//
//   Capture raw data from p5 glove, normalize to [0., 1.].
//   Normalize again using user specific values that span physical
//     use of the glove.
//
//   USER CUSTOMIZED values defined as constants of the p5Glove object. 
//
//   (commented) Option to create averages in min/max; use adaptive
//     algorithm to adjust min/max per recent history of use.
//
//
//   XXX  WARNING -- sometimes wrist values start out as NaN and do
//                   not recover.
// 	  WORKAROUND -- stop and restart glove.
//   
//
// INLETS--
//   0: list representing payload from a single udp packet
// 
// MESSAGES IN--
//   bang 			dump controller values
//   reset 			recreate glove object to reset state
//   udpaccept [toggle] 	enable/disable UDP packet processing
//   uservalues 		dump user customized settings
//   (receive) p5udpinput_input
//
// OUTLETS--
//   (see below)
// 
// PUBLIC FUNCTIONS--
//   experiment(), anything(), bang()
//   reset(), udpaccept()
//   p5Glove.*
//     update()
// 
// PRIVATE FUNCTIONS--
//   dumpFingerWrist()
//     *.mkmsg()
//   dumpUserCustom()
//   g5Glove.*
//     initGloveElement(), elementUpdate()
//

//							--o-
// MAX OBJECT MANAGEMENT
//

SCRIPTNAME = "p5udpinput";

autowatch = 1;		// recompile when editor saves file 
debug = 1;              // global debug: higher value for greater verbosity

inlets = 1;
setinletassist(0, "(list) UDP data; bang to dump glove values");

outlets = 13;
setoutletassist(0, "(bool) buttonA");
setoutletassist(1, "(bool) buttonB");
setoutletassist(2, "(bool) buttonC");

setoutletassist(3, "(float) thumb");
setoutletassist(4, "(float) index finger");
setoutletassist(5, "(float) middle finger");
setoutletassist(6, "(float) ring finger");
setoutletassist(7, "(float) pinky finger");

setoutletassist(8, "(float) yaw");
setoutletassist(9, "(float) pitch");
setoutletassist(10, "(float) roll");

setoutletassist(11, "(list) data on all continuous controllers");
setoutletassist(12, "NOT USED");



//							--o-
// BUILT-IN MESSAGES
//

function bang() 
{ 
  dumpFingerWrist(gloveObjP); 
}


//----------------------------------- -o-
function anything()
{
  var debugstr 	= ds = "anything";
  var args 	= arrayfromargs(arguments);
  var instr 	= "inlet(" + inlet + ")  " + messagename + ": " + args;

  // if (debug) { db(ds); }


  if (gloveObjP == undefined) { 
    gloveObjP = new p5Glove();
  }


  //
  // parse incoming message
  //
  if (messagename != gloveObjP.udpTag) 
  {
    switch (messagename) 
    {
    case "udpaccept 0":
      gloveObjP.udpAccept = false;
      infomsg("DISABLED : Receipt of UDP packets.");
      break;

    case "udpaccept 1":
      gloveObjP.udpAccept = true;
      infomsg("ENABLED : Receipt of UDP packets.");
      break;

    default:
      
      if (messagename == gloveObjP.udpTag) { 
        gloveObjP.udpAccept = true; 
	infomsg("ENABLED : Receipt of UDP packets.");
      }

      else 
      {
	if (debug) { dx(ds, instr); }
	return;
      }

    }  // switch()
  }



  //
  // handle incoming UDP packets, outlet the results
  //
  if (! gloveObjP.udpAccept) { return; }

  gloveObjP.update( args[0], args[1], args[2], 			// buttons
  			args[3], args[4],			// fingers
  			args[5], args[6], args[7],
			  args[8], args[9], args[10] ); 	// wrist


  outlet(0, gloveObjP.buttonA); 		// on/off == 1/0
  outlet(1, gloveObjP.buttonB);
  outlet(2, gloveObjP.buttonC);

  // NB  .normUser instead of .normalized
  //
  outlet(3, 1.0 - gloveObjP.thumb.normUser); 		// in/out == [0., 1.]
  outlet(4, 1.0 - gloveObjP.index.normUser);
  outlet(5, 1.0 - gloveObjP.middle.normUser);
  outlet(6, 1.0 - gloveObjP.ring.normUser);
  outlet(7, 1.0 - gloveObjP.pinky.normUser);

  outlet(8, gloveObjP.yaw.normUser); 		// center/side == [0., 1.]
  outlet(9, gloveObjP.pitch.normUser); 		// down/up == [0., 1.]
  outlet(10, 1.0 - gloveObjP.roll.normUser); 	// center/right == [0., 1.]


  // if (debug) { de(ds); }

}  // anything()




//							--o-
// USER OBJECTS
//

var gloveObjP = undefined;

function p5Glove() 					// --o-
{
  // ds = "p5Glove"; 


  // CONSTANTS 						-o-
  //
  this.udpTag 		= "/p5glove_data";
  this.udpAccept 	= 0; 	// 1/0 == enable/disable UDP processing


  // how often to store (sample) incoming input into value arrays
  //
  updateFrequency 	= 3;		// 0 = update from each UDP packet
  updateFrequencyCount 	= 0;
  
  NOVALUE = -1.0;

    		// ------------------------------------------------------
  		// XXX !!!!!!  USER CUSTOMIZATION  (BEGIN)

  this.THUMBLOW 	= NOVALUE; 	// TBD  externalize user values
  this.THUMBHIGH 	= NOVALUE;
  this.THUMBJITTER 	= NOVALUE;

  this.INDEXLOW 	= NOVALUE;
  this.INDEXHIGH 	= NOVALUE;
  this.INDEXJITTER 	= NOVALUE;

  this.MIDDLELOW 	= NOVALUE;
  this.MIDDLEHIGH 	= NOVALUE;
  this.MIDDLEJITTER 	= NOVALUE;

  this.RINGLOW 		= NOVALUE;
  this.RINGHIGH 	= NOVALUE;
  this.RINGJITTER 	= NOVALUE;

  this.PINKYLOW 	= NOVALUE;
  this.PINKYHIGH 	= NOVALUE;
  this.PINKYJITTER 	= NOVALUE;


  this.YAWLOW 		= 0.25;
  this.YAWHIGH 		= 0.60;
  this.YAWJITTER 	= 0.90;
  	// NB  forward is zero; right/left increase to 1.0

  this.PITCHLOW 	= 0.08;
  this.PITCHHIGH 	= 0.80;
  this.PITCHJITTER 	= 0.75;

  this.ROLLLOW 		= 0.05;
  this.ROLLHIGH 	= 0.43;
  this.ROLLJITTER 	= 0.75;
  	// NB  assume mid is zero and right increases to 1.0;
	//	 this because of hand position, not values returned
	//	 from the glove


  		// XXX !!!!!!  USER CUSTOMIZATION  (END)
    		// ------------------------------------------------------


  // VARIABLES 						-o-
  //
  this.buttonA = false;
  this.buttonB = false;
  this.buttonC = false;

  this.thumb = new initGloveElement(
		"thumb", this.THUMBLOW, this.THUMBHIGH, this.THUMBJITTER );
  this.index = new initGloveElement(
		"index", this.INDEXLOW, this.INDEXHIGH, this.INDEXJITTER );
  this.middle = new initGloveElement(
		"middle", this.MIDDLELOW, this.MIDDLEHIGH, this.MIDDLEJITTER );
  this.ring = new initGloveElement(
		"ring", this.RINGLOW, this.RINGHIGH, this.RINGJITTER );
  this.pinky = new initGloveElement(
		"pinky", this.PINKYLOW, this.PINKYHIGH, this.PINKYJITTER );

  this.yaw = new initGloveElement(
		"yaw", this.YAWLOW, this.YAWHIGH, this.YAWJITTER );
  this.pitch = new initGloveElement(
		"pitch", this.PITCHLOW, this.PITCHHIGH, this.PITCHJITTER );
  this.roll = new initGloveElement(
		"roll", this.ROLLLOW, this.ROLLHIGH, this.ROLLJITTER );



  // PRIVATE FUNCTIONS 					-o-
  //
  function initGloveElement(controllerType, NORMLOW, NORMHIGH, NORMJITTER)
  { 
    // var ds = "p5Glove.initGloveElement";


    this.init 		= false;
    this.initCount 	= 0; 		
      // process initCount UDP packets before activating the object
    this.name 		= controllerType;

    this.min 		= 0.0;
    // this.minArray 	= undefined;	// NB  smooth min by average
    // this.minArrayLen = 11;

    this.max 		= 0.0;
    // this.maxArray 	= undefined; 	// NB  smooth max by average
    // this.maxArrayLen = 11;

    this.current 		= 0.0;
    this.currentArray 		= undefined;
    this.currentArrayLen 	= 11;
      // average over last currentArrayLen values from UDP stream 

    this.normalized 	= -1.0;


    		// ------------------------------------------------------
  		// XXX !!!!!!  USER CUSTOMIZATION  (BEGIN)

    // NB  min and max are customed to a given enduser's habits
    //     within the ranges provided by this code module.
    //
    //     jitter discards values out of normUser radius
    //
    this.normUser 	= 0.0;
    this.normUserPrev 	= 0.0;

    this.normUserLow 	= NORMLOW; 	
    this.normUserHigh 	= NORMHIGH; 
    this.normUserJitter = NORMJITTER;

  		// XXX !!!!!!  USER CUSTOMIZATION  (END)
    		// ------------------------------------------------------


    this.isFinger 	= true; 		// false == "isWrist"

    this.range 		= 1.0;
    // this.rangeDivisor 	= 12;	// NB  (cf. below)
    	// min/max buffer allowing recalibration to apparent range of motion;
	//  rangeDivisor inversely proportional to buffer size
    // this.rangeJitter 	= 0;	// NB  (cf. below)
    	// min/max changes greater than this are discarded as noise


    switch (controllerType)
    {
    case "thumb": 					// finger controllers
    case "index": 	// NB  customize divisor, arrays, jitter... 
    case "middle":
    case "ring":
    case "pinky":
      /*EMPTY*/
      break;

    case "yaw": 					// wrist controllers
    case "pitch": 	// NB  customize divisor, arrays, jitter...
    case "roll":
      this.isFinger = false;
      break;


    default:
      errmsg(ds + "()" + QCOL + 
        "unrecognized controllerType \"" + controllerType + "\"");

    }  // switch(controllerType)


    // this.minArray 	= new Array(this.minArrayLen); 	// NB  heuristic
    // this.maxArray 	= new Array(this.maxArrayLen); 	// NB  heuristic
    this.currentArray 	= new Array(this.currentArrayLen);

    // zeroArray(this.maxArray); 			// NB  heuristic
    // zeroArray(this.minArray); 			// NB  heuristic
    zeroArray(this.currentArray);

  }  // initGloveElement()


  // elementUpdate()
  //
  //   FIX  Check for NaN's before updating curAveArray
  //
  function elementUpdate(element, currentValue)
  {
    // var ds = "p5Glove.elementUpdate"


    element.current = curAveArray(element.currentArray, currentValue);

    if (! element.init) 
    {
      if (element.initCount++ < element.currentArrayLen) { return; }

      element.min = element.max = element.current;
      element.init = true;

      return;
    }


    //  update element.{min,max} --
    //    simple: let min/max expand to and hold peak points.
    //
    //			OR: heuristic...
    //    . ignore differences greater than element.rangeJitter
    //    . update when element.current is within element.rangeDivisor;
    //        this heuristic increases/shrinks min/max as the
    //        range of the controller shifts and changes shape
    //

    if ( element.current < element.min )		// MINIMUM
    {
	element.min = element.current;
    }

    /* // ----------------------------------- 
    //
    //  NB  (disabled) heuristic for range adaption (min)
    //
    if ( element.current 
	    < (element.min + (element.range / element.rangeDivisor)) ) 
    {
      if ( element.rangeJitter <= Math.abs(element.min - element.current) )
      { 					// NB  heuristic
	element.min = element.current;
      }
    }
    */ // ----------------------------------- 


    if ( element.current > element.max )		// MAXIMUM
    {
	element.max = element.current;
    }

    /* // ----------------------------------- 
    //
    //  NB  (disabled) heuristic for range adaption (max)
    //
    if ( element.current 
	    > (element.max - (element.range / element.rangeDivisor)) ) 
    {
      if ( element.rangeJitter <= Math.abs(element.max - element.current) )
      {
	element.max = element.current;
      } 					// NB
    }
    */ // ----------------------------------- 


    element.range = Math.abs(element.max - element.min);


    // normalize position from glove data:  [min, max] -> [0., 1.]
    //   TBD  compensate for tessitura of use?  how to make history?
    //
    element.normalized = 1. *  
    	fixOutOfBounds( (element.current - element.min) / (element.range) );


    //
    // normalize according to personal end user values against
    //   normalized value from glove data
    //
    		// ------------------------------------------------------
  		// XXX !!!!!!  USER CUSTOMIZATION  (BEGIN)

    if ( (element.normUserLow == -1.0) || (element.normUserHigh == -1.0) )
    {
      element.normUser = element.normalized;
      return;
    }

    // ignore jumps out of range or outside jitter radius
    //
    //   XXX  should jitter change per segmented range of normUser?
    //
    if ( (element.normalized < element.normUserLow)
	      || (element.normalized > element.normUserHigh) 
	            || (element.normalized == 0)
	)
    {
      element.normUser = element.normUserPrev;
      return;
    }

    // FIX  jitter check fails around controler output containing an
    //        origin of zero
    //
    if ( element.normUserJitter 
	  < Math.abs(element.normUserPrev - element.normUser) )
    {
      return;
    }


    // normalize again for enduser gestural range
    //
    element.normUserPrev = element.normUser;

    element.normUser = 1.0 *
	fixOutOfBounds( 
	      (element.normalized - element.normUserLow)
		/ (element.normUserHigh - element.normUserLow)
	);

  		// XXX !!!!!!  USER CUSTOMIZATION  (END)
    		// ------------------------------------------------------
    
  }  // elementUpdate()
 


  // PUBLIC FUNCTIONS 					-o-
  //
  this.update = function( buttonA, buttonB, buttonC, 
			    thumb, index, middle, ring, pinky, 
			      yaw, pitch, roll )
  {
    // var ds = "p5Glove.update";


    // push current value once every updateFrequency cycles
    //
    if ( (updateFrequencyCount++ < updateFrequency) ) { return; }

    updateFrequencyCount = 0;

    this.buttonA = buttonA;
    this.buttonB = buttonB;
    this.buttonC = buttonC;

    elementUpdate( this.thumb, thumb );
    elementUpdate( this.index, index );
    elementUpdate( this.middle, middle );
    elementUpdate( this.ring, ring );
    elementUpdate( this.pinky, pinky );
    
    elementUpdate( this.yaw, yaw );
    elementUpdate( this.pitch, pitch );
    elementUpdate( this.roll, roll );

  }  // p5Glove.update()


  //  Dump values customized for enduser habits of motion.
  //
  this.dumpCustomValues = function()
  {
    pl(); pl(); pl(); ps();

    infomsgl("thumb", 
      this.THUMBLOW + QSLASH + this.THUMBHIGH + QSLASH + this.THUMBJITTER);
    infomsgl("index", 
      this.INDEXLOW + QSLASH + this.INDEXHIGH + QSLASH + this.INDEXJITTER);
    infomsgl("middle", 
      this.MIDDLELOW + QSLASH + this.MIDDLEHIGH + QSLASH + this.MIDDLEJITTER);
    infomsgl("ring", 
      this.RINGLOW + QSLASH + this.RINGHIGH + QSLASH + this.RINGJITTER);
    infomsgl("pinky", 
      this.PINKYLOW + QSLASH + this.PINKYHIGH + QSLASH + this.PINKYJITTER);

    ps();

    infomsgl("yaw", 
      this.YAWLOW + QSLASH + this.YAWHIGH + QSLASH + this.YAWJITTER);
    infomsgl("pitch", 
      this.PITCHLOW + QSLASH + this.PITCHHIGH + QSLASH + this.PITCHJITTER);
    infomsgl("roll", 
      this.ROLLLOW + QSLASH + this.ROLLHIGH + QSLASH + this.ROLLJITTER);

    pl(); pl(); pl(); ps();

  }  // p5Glove.dumpCustomValues()

}  // p5Glove()




//							--o-
// USER FUNCTIONS
//

//----------------------------------- -o-
// dumpFingerWrist()
//
//   Post information naming each continuous controllers followed by 
//     <min> (<current>) <max>.
//
// OUTLETS--
//   11: (string) controller values
//
dumpFingerWrist.local = 1;

function dumpFingerWrist(gp)
{

  mkmsg.local = 1;
  function mkmsg(s, label, element)
  {
    return(
	s + label
	  + "(" + roundFloat(element.normalized, 1000) 
	  + ")   " 
	    + "\t   range   "
	      + roundFloat(element.range, 10)

				  /* 			// NB  heuristic
	        + "  (d: " + element.rangeDivisor 
	        + ", j: " + element.rangeJitter + ")"
				  */ 			// NB  heuristic
				  /* 			// NB  heuristic
	    + "\t   deltas   "
	      + roundFloat(element.minDelta, 10)
	        + QCOL + roundFloat(element.maxDelta, 10)
				  */ 			// NB  heuristic


	    + "\t    controller   "
	      + roundFloat(element.min, 10) + SP 
	        + "(" + roundFloat(element.current, 10) + ")" + SP 
		 + roundFloat(element.max, 10)


	    + SP + SP + SP + SP + QDASH + SP + SP + SP + SP 
		  + "low/high/jitter" + QCOL 
		      + element.normUserLow + SP + SLASH + SP 
		        + element.normUserHigh + SP + SLASH + SP 
			  + element.normUserJitter
	      + SP + SP + SP
		  + "user/prev" + QCOL 
		      + roundFloat(element.normUser, 1000) + SP + SLASH + SP
			+ roundFloat(element.normUserPrev, 1000)
	  + CR
    );
  }


  pl(); ps();

  outlet( 11, mkmsg("", " THUMB", gp.thumb) );
  outlet( 11, mkmsg("", "     INDEX", gp.index) );
  outlet( 11, mkmsg("", "     MIDDLE", gp.middle) );
  outlet( 11, mkmsg("", "     RING", gp.ring) );
  outlet( 11, mkmsg("", "     PINKY", gp.pinky) );

  ps();

  outlet( 11, mkmsg("", " YAW", gp.yaw) );
  outlet( 11, mkmsg("", "     PITCH", gp.pitch) );
  outlet( 11, mkmsg("", "     ROLL", gp.roll) );

}  // dumpFingerWrist()



// 
// dump only user customized values
//

function uservalues() 
{
  gloveObjP.dumpCustomValues();
}  



//----------------------------------- -o-
// reset()
//
//  Zero all glove values.  
//
//  Equivalent to deleting and recreating p5Glove object.  XXX
//

function reset()
{
  delete (gloveObjP);
  gloveObjP = new p5Glove();

  infomsg("Reset all UDP processing values.");  

}  // reset()




//							--o-
// TESTS AND EXEPERIMENTS
//

//----------------------------------- -o-
// experiment()
//
// Something risque.
//
function experiment()
{
  var debugstr 	= ds = "experiment";
  var args 	= arrayfromargs(arguments);

  pl(); ps();
  if (debug) { db(ds); da(ds, "args", args); }


  if (false) {	// what is infinity?
    var big = Infinity;
    var small = -Infinity;

    dlx(ds, "big", big);
    dlx(ds, "small", small);

    dlx(ds, "is big infinity? ", big == Infinity);
    dlx(ds, "is big -infinity? ", big == -Infinity);

  }


  if (false) {	// test roundoff
    var big = 12.3456789567;
    var format = 1.00;

    var outout = Math.round( big * 100 ) / 100;


    dx(ds, roundFloat(big, 1000));
  }


  if (false) {	// test Array size and zero out
    var alen = 5
    var a = new Array(alen);

    dla(ds, "a", a);

    zeroArray(a, alen);

    dla(ds, "a", a);
  }


  if (false) {		// WAS  test array.shift() and .push()
  			// test curAveArray()
    var a = new Array();
    var alen = 5;

    for (var i = 0; i <= 10; i++) 
    {
      // var curave = curAveArray(a, i, alen);	// BROKEN
      dla(ds, "a", a);
      // dlx( ds, "average", curAveArray(a, i, alen) );	// BROKEN
    }
  }


  if (debug) { de(ds); }

}  // experiment()

